#!/usr/bin/env python

__version__ = '$Revision: 4791 $'.split()[1]
__date__ = '$Date: 2007-02-05 $'.split()[1]
__author__ = 'xmlbinmsg'

__doc__='''

Autogenerated python functions to serialize/deserialize binary messages.

Generated by: ./aisxmlbinmsg2py.py

Need to then wrap these functions with the outer AIS packet and then
convert the whole binary blob to a NMEA string.  Those functions are
not currently provided in this file.

serialize: python to ais binary
deserialize: ais binary to python

The generated code uses translators.py, binary.py, and aisstring.py
which should be packaged with the resulting files.


@requires: U{epydoc<http://epydoc.sourceforge.net/>} > 3.0alpha3
@requires: U{BitVector<http://cheeseshop.python.org/pypi/BitVector>}

@author: '''+__author__+'''
@version: ''' + __version__ +'''
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ parser
@status: under development
@license: Generated code has no license
'''

import sys
from decimal import Decimal
from BitVector import BitVector

import binary, aisstring

# FIX: check to see if these will be needed
TrueBV  = BitVector(bitstring="1")
"Why always rebuild the True bit?  This should speed things up a bunch"
FalseBV = BitVector(bitstring="0")
"Why always rebuild the False bit?  This should speed things up a bunch"


fieldList = [
	'time_month',
	'time_day',
	'time_hour',
	'time_min',
	'stationid',
	'pos_longitude',
	'pos_latitude',
	'flow',
	'reserved',
]

def encode(params, validate=False):
	'''Create a sls_wind binary message payload to pack into an AIS Msg sls_wind.

	Fields in params:
	  - time_month(uint): Time tag of measurement  month 1..12
	  - time_day(uint): Time tag of measurement  day of the month 1..31
	  - time_hour(uint): Time tag of measurement  UTC hours 0..23
	  - time_min(uint): Time tag of measurement  minutes
	  - stationid(aisstr6): Character identifier of the station
	  - pos_longitude(decimal): Location of measurement  East West location
	  - pos_latitude(decimal): Location of measurement  North South location
	  - flow(uint): Water flow
	  - reserved(uint): Reserved bits for future use (field automatically set to "0")
	@param params: Dictionary of field names/values.  Throws a ValueError exception if required is missing
	@param validate: Set to true to cause checking to occur.  Runs slower.  FIX: not implemented.
	@rtype: BitVector
	@return: encoded binary message (for binary messages, this needs to be wrapped in a msg 8
	@note: The returned bits may not be 6 bit aligned.  It is up to you to pad out the bits.
	'''

	bvList = []
	bvList.append(binary.setBitVectorSize(BitVector(intVal=params['time_month']),4))
	bvList.append(binary.setBitVectorSize(BitVector(intVal=params['time_day']),5))
	bvList.append(binary.setBitVectorSize(BitVector(intVal=params['time_hour']),5))
	bvList.append(binary.setBitVectorSize(BitVector(intVal=params['time_min']),6))
	if 'stationid' in params:
		bvList.append(aisstring.encode(params['stationid'],42))
	else:
		bvList.append(aisstring.encode('@@@@@@@',42))
	if 'pos_longitude' in params:
		bvList.append(binary.bvFromSignedInt(int(Decimal(params['pos_longitude'])*Decimal('60000')),25))
	else:
		bvList.append(binary.bvFromSignedInt(10860000,25))
	if 'pos_latitude' in params:
		bvList.append(binary.bvFromSignedInt(int(Decimal(params['pos_latitude'])*Decimal('60000')),24))
	else:
		bvList.append(binary.bvFromSignedInt(5460000,24))
	if 'flow' in params:
		bvList.append(binary.setBitVectorSize(BitVector(intVal=params['flow']),10))
	else:
		bvList.append(binary.setBitVectorSize(BitVector(intVal=16383),10))
	bvList.append(binary.setBitVectorSize(BitVector(intVal=0),19))

	return binary.joinBV(bvList)

def decode(bv, validate=False):
	'''Unpack a sls_wind message 

	Fields in params:
	  - time_month(uint): Time tag of measurement  month 1..12
	  - time_day(uint): Time tag of measurement  day of the month 1..31
	  - time_hour(uint): Time tag of measurement  UTC hours 0..23
	  - time_min(uint): Time tag of measurement  minutes
	  - stationid(aisstr6): Character identifier of the station
	  - pos_longitude(decimal): Location of measurement  East West location
	  - pos_latitude(decimal): Location of measurement  North South location
	  - flow(uint): Water flow
	  - reserved(uint): Reserved bits for future use (field automatically set to "0")
	@type bv: BitVector
	@param bv: Bits defining a message
	@param validate: Set to true to cause checking to occur.  Runs slower.  FIX: not implemented.
	@rtype: dict
	@return: params
	'''

	#Would be nice to check the bit count here..
	#if validate:
	#	assert (len(bv)==FIX: SOME NUMBER)
	r = {}
	r['time_month']=int(bv[0:4])
	r['time_day']=int(bv[4:9])
	r['time_hour']=int(bv[9:14])
	r['time_min']=int(bv[14:20])
	r['stationid']=aisstring.decode(bv[20:62])
	r['pos_longitude']=Decimal(binary.signedIntFromBV(bv[62:87]))/Decimal('60000')
	r['pos_latitude']=Decimal(binary.signedIntFromBV(bv[87:111]))/Decimal('60000')
	r['flow']=int(bv[111:121])
	r['reserved']=0
	return r

def decodetime_month(bv, validate=False):
	return int(bv[0:4])

def decodetime_day(bv, validate=False):
	return int(bv[4:9])

def decodetime_hour(bv, validate=False):
	return int(bv[9:14])

def decodetime_min(bv, validate=False):
	return int(bv[14:20])

def decodestationid(bv, validate=False):
	return aisstring.decode(bv[20:62])

def decodepos_longitude(bv, validate=False):
	return Decimal(binary.signedIntFromBV(bv[62:87]))/Decimal('60000')

def decodepos_latitude(bv, validate=False):
	return Decimal(binary.signedIntFromBV(bv[87:111]))/Decimal('60000')

def decodeflow(bv, validate=False):
	return int(bv[111:121])

def decodereserved(bv, validate=False):
	return 0


def printHtml(params, out=sys.stdout):
		out.write("<h3>sls_wind<h3>\n")
		out.write("<table border=\"1\">\n")
		out.write("<tr bgcolor=\"orange\">\n")
		out.write("<th align=\"left\">Field Name</th>\n")
		out.write("<th align=\"left\">Type</th>\n")
		out.write("<th align=\"left\">Value</th>\n")
		out.write("<th align=\"left\">Value in Lookup Table</th>\n")
		out.write("<th align=\"left\">Units</th>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>time_month</td>\n")
		out.write("<td>uint</td>\n")
		if 'time_month' in params:
			out.write("	<td>"+str(params['time_month'])+"</td>\n")
			out.write("	<td>"+str(params['time_month'])+"</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>time_day</td>\n")
		out.write("<td>uint</td>\n")
		if 'time_day' in params:
			out.write("	<td>"+str(params['time_day'])+"</td>\n")
			out.write("	<td>"+str(params['time_day'])+"</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>time_hour</td>\n")
		out.write("<td>uint</td>\n")
		if 'time_hour' in params:
			out.write("	<td>"+str(params['time_hour'])+"</td>\n")
			out.write("	<td>"+str(params['time_hour'])+"</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>time_min</td>\n")
		out.write("<td>uint</td>\n")
		if 'time_min' in params:
			out.write("	<td>"+str(params['time_min'])+"</td>\n")
			out.write("	<td>"+str(params['time_min'])+"</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>stationid</td>\n")
		out.write("<td>aisstr6</td>\n")
		if 'stationid' in params:
			out.write("	<td>"+str(params['stationid'])+"</td>\n")
			out.write("	<td>"+str(params['stationid'])+"</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>pos_longitude</td>\n")
		out.write("<td>decimal</td>\n")
		if 'pos_longitude' in params:
			out.write("	<td>"+str(params['pos_longitude'])+"</td>\n")
			out.write("	<td>"+str(params['pos_longitude'])+"</td>\n")
		out.write("<td>degrees</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>pos_latitude</td>\n")
		out.write("<td>decimal</td>\n")
		if 'pos_latitude' in params:
			out.write("	<td>"+str(params['pos_latitude'])+"</td>\n")
			out.write("	<td>"+str(params['pos_latitude'])+"</td>\n")
		out.write("<td>degrees</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>flow</td>\n")
		out.write("<td>uint</td>\n")
		if 'flow' in params:
			out.write("	<td>"+str(params['flow'])+"</td>\n")
			out.write("	<td>"+str(params['flow'])+"</td>\n")
		out.write("<td>m^3/s</td>\n")
		out.write("</tr>\n")
		out.write("\n")
		out.write("<tr>\n")
		out.write("<td>reserved</td>\n")
		out.write("<td>uint</td>\n")
		if 'reserved' in params:
			out.write("	<td>"+str(params['reserved'])+"</td>\n")
			out.write("	<td>"+str(params['reserved'])+"</td>\n")
		out.write("</tr>\n")
		out.write("</table>\n")


def printKml(params, out=sys.stdout):
	'''KML (Keyhole Markup Language) for Google Earth, but without the header/footer'''
	out.write("\	<Placemark>\n")
	out.write("\t	<name>"+str(params['stationid'])+"</name>\n")
	out.write("\t\t<description>\n")
	import StringIO
	buf = StringIO.StringIO()
	printHtml(params,buf)
	import cgi
	out.write(cgi.escape(buf.getvalue()))
	out.write("\t\t</description>\n")
	out.write("\t\t<styleUrl>#m_ylw-pushpin_copy0</styleUrl>\n")
	out.write("\t\t<Point>\n")
	out.write("\t\t\t<coordinates>")
	out.write(str(params['pos_longitude']))
	out.write(',')
	out.write(str(params['pos_latitude']))
	out.write(",0</coordinates>\n")
	out.write("\t\t</Point>\n")
	out.write("\t</Placemark>\n")

def printFields(params, out=sys.stdout, format='std', fieldList=None):
	'''Print a reserved message to stdout.

	Fields in params:
	  - time_month(uint): Time tag of measurement  month 1..12
	  - time_day(uint): Time tag of measurement  day of the month 1..31
	  - time_hour(uint): Time tag of measurement  UTC hours 0..23
	  - time_min(uint): Time tag of measurement  minutes
	  - stationid(aisstr6): Character identifier of the station
	  - pos_longitude(decimal): Location of measurement  East West location
	  - pos_latitude(decimal): Location of measurement  North South location
	  - flow(uint): Water flow
	  - reserved(uint): Reserved bits for future use (field automatically set to "0")
	@param params: Dictionary of field names/values.  
	@param out: File like object to write to
	@rtype: stdout
	@return: text to out
	'''

	if 'std'==format:
		out.write("reserved:\n")
		if 'time_month' in params: out.write("	time_month:     "+str(params['time_month'])+"\n")
		if 'time_day' in params: out.write("	time_day:       "+str(params['time_day'])+"\n")
		if 'time_hour' in params: out.write("	time_hour:      "+str(params['time_hour'])+"\n")
		if 'time_min' in params: out.write("	time_min:       "+str(params['time_min'])+"\n")
		if 'stationid' in params: out.write("	stationid:      "+str(params['stationid'])+"\n")
		if 'pos_longitude' in params: out.write("	pos_longitude:  "+str(params['pos_longitude'])+"\n")
		if 'pos_latitude' in params: out.write("	pos_latitude:   "+str(params['pos_latitude'])+"\n")
		if 'flow' in params: out.write("	flow:           "+str(params['flow'])+"\n")
		if 'reserved' in params: out.write("	reserved:       "+str(params['reserved'])+"\n")
	elif 'csv'==format:
		if None == options.fieldList:
			options.fieldList = fieldList
		needComma = False;
		for field in fieldList:
			if needComma: out.write(',')
			needComma = True
			if field in params:
				out.write(str(params[field]))
			# else: leave it empty
		out.write("\n")
	elif 'html'==format:
		printHtml(params,out)
	elif 'sql'==format:
		sqlInsertStr(params,out)
	elif 'kml'==format:
		printKml(params,out)
	elif 'kml-full'==format:
		out.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n")
		out.write("<kml xmlns=\"http://earth.google.com/kml/2.1\">\n")
		out.write("<Document>\n")
		out.write("	<name>sls_wind</name>\n")
		printKml(params,out)
		out.write("</Document>\n")
		out.write("</kml>\n")
	else: 
		print "ERROR: unknown format:",format
		assert False

	return # Nothing to return

######################################################################
# SQL SUPPORT
######################################################################

def sqlCreateStr(outfile=sys.stdout, fields=None, extraFields=None):
	'''
	Return the SQL CREATE command for this message type
	@param outfile: file like object to print to.
	@param fields: which fields to put in the create.  Defaults to all.
	@param extraFields: A sequence of tuples containing (name,sql type) for additional fields
	@return: sql create string
	@rtype: str

	@see: sqlCreate
	'''
	outfile.write(str(sqlCreate(fields,extraFields)))

def sqlCreate(fields=None, extraFields=None):
	'''
	Return the sqlhelp object to create the table.

	@param fields: which fields to put in the create.  Defaults to all.
	@param extraFields: A sequence of tuples containing (name,sql type) for additional fields
	@return: An object that can be used to generate a return
	@rtype: sqlhelp.create
	'''
	if None == fields: fields = fieldList
	import sqlhelp
	c = sqlhelp.create('sls_wind')
	if 'time_month' in fields: c.addInt ('time_month')
	if 'time_day' in fields: c.addInt ('time_day')
	if 'time_hour' in fields: c.addInt ('time_hour')
	if 'time_min' in fields: c.addInt ('time_min')
	if 'stationid' in fields: c.addVarChar(stationid,7)
	if 'pos_longitude' in fields: c.addDecimal('pos_longitude',7,4)
	if 'pos_latitude' in fields: c.addDecimal('pos_latitude',7,4)
	if 'flow' in fields: c.addInt ('flow')
	if 'reserved' in fields: c.addInt ('reserved')

	return c

def sqlInsertStr(params, outfile=sys.stdout, extraParams=None):
	'''
	Return the SQL CREATE command for this message type
	@param params: dictionary of values keyed by field name
	@param outfile: file like object to print to.
	@param extraParams: A sequence of tuples containing (name,sql type) for additional fields
	@return: sql create string
	@rtype: str

	@see: sqlCreate
	'''
	outfile.write(str(sqlInsert(params,extraParams)))


def sqlInsert(params,extraParams=None):
	'''
	Give the SQL insert statement
	@param params: dict keyed by field name of values
	@param extraParams: any extra fields that you have created beyond the normal ais message fields
	@rtype: sqlhelp.insert
	@return: insert class instance
	@todo: allow optional type checking of params?
	@warning: this will take invalid keys happily and do what???
	'''
	import sqlhelp
	i = sqlhelp.insert('sls_wind')
	for key in params: i.add(key,params[key])
	if None != extraParams:
		for key in extraParams: 
			i.add(key,extraParams[key])

	return i


######################################################################
# UNIT TESTING
######################################################################
import unittest
def testParams():
	'''Return a params file base on the testvalue tags.
	@rtype: dict
	@return: params based on testvalue tags
	'''
	params = {}
	params['time_month'] = 2
	params['time_day'] = 28
	params['time_hour'] = 23
	params['time_min'] = 45
	params['stationid'] = 'A345678'
	params['pos_longitude'] = Decimal('-122.16328')
	params['pos_latitude'] = Decimal('37.42446')
	params['flow'] = 43
	params['reserved'] = 0

	return params

class Testsls_wind(unittest.TestCase):
	'''Use testvalue tag text from each type to build test case the sls_wind message'''
	def testEncodeDecode(self):

		params = testParams()
		bits   = encode(params)
		r      = decode(bits)

		# Check that each parameter came through ok.
		self.failUnlessEqual(r['time_month'],params['time_month'])
		self.failUnlessEqual(r['time_day'],params['time_day'])
		self.failUnlessEqual(r['time_hour'],params['time_hour'])
		self.failUnlessEqual(r['time_min'],params['time_min'])
		self.failUnlessEqual(r['stationid'],params['stationid'])
		self.failUnlessAlmostEqual(r['pos_longitude'],params['pos_longitude'],4)
		self.failUnlessAlmostEqual(r['pos_latitude'],params['pos_latitude'],4)
		self.failUnlessEqual(r['flow'],params['flow'])
		self.failUnlessEqual(r['reserved'],params['reserved'])

def addMsgOptions(parser):
	parser.add_option('-d','--decode',dest='doDecode',default=False,action='store_true',
		help='decode a "sls_wind" AIS message')
	parser.add_option('-e','--encode',dest='doEncode',default=False,action='store_true',
		help='encode a "sls_wind" AIS message')
	parser.add_option('--time_month-field', dest='time_monthField',metavar='uint',type='int'
		,help='Field parameter value [default: %default]')
	parser.add_option('--time_day-field', dest='time_dayField',metavar='uint',type='int'
		,help='Field parameter value [default: %default]')
	parser.add_option('--time_hour-field', dest='time_hourField',metavar='uint',type='int'
		,help='Field parameter value [default: %default]')
	parser.add_option('--time_min-field', dest='time_minField',metavar='uint',type='int'
		,help='Field parameter value [default: %default]')
	parser.add_option('--stationid-field', dest='stationidField',default='@@@@@@@',metavar='aisstr6',type='string'
		,help='Field parameter value [default: %default]')
	parser.add_option('--pos_longitude-field', dest='pos_longitudeField',default=Decimal('181'),metavar='decimal',type='string'
		,help='Field parameter value [default: %default]')
	parser.add_option('--pos_latitude-field', dest='pos_latitudeField',default=Decimal('91'),metavar='decimal',type='string'
		,help='Field parameter value [default: %default]')
	parser.add_option('--flow-field', dest='flowField',default=16383,metavar='uint',type='int'
		,help='Field parameter value [default: %default]')

############################################################
if __name__=='__main__':

	from optparse import OptionParser
	parser = OptionParser(usage="%prog [options]",
		version="%prog "+__version__)

	parser.add_option('--doc-test',dest='doctest',default=False,action='store_true',
		help='run the documentation tests')
	parser.add_option('--unit-test',dest='unittest',default=False,action='store_true',
		help='run the unit tests')
	parser.add_option('-v','--verbose',dest='verbose',default=False,action='store_true',
		help='Make the test output verbose')

	# FIX: remove nmea from binary messages.  No way to build the whole packet?
	# FIX: or build the surrounding msg 8 for a broadcast?
	typeChoices = ('binary','nmeapayload','nmea') # FIX: what about a USCG type message?
	parser.add_option('-t','--type',choices=typeChoices,type='choice',dest='ioType'
		,default='nmeapayload'
		,help='What kind of string to expect ('+', '.join(typeChoices)+') [default: %default]')


	outputChoices = ('std','html','csv','sql' , 'kml','kml-full')
	parser.add_option('-T','--output-type',choices=outputChoices,type='choice',dest='outputType'
		,default='std'
		,help='What kind of string to output ('+', '.join(outputChoices)+') [default: %default]')

	parser.add_option('-o','--output',dest='outputFileName',default=None,
			  help='Name of the python file to write [default: stdout]')

	parser.add_option('-f','--fields',dest='fieldList',default=None, action='append',
			  choices=fieldList,
			  help='Which fields to include in the output.  Currently only for csv output [default: all]')

	parser.add_option('-p','--print-csv-field-list',dest='printCsvfieldList',default=False,action='store_true',
			  help='Print the field name for csv')

	parser.add_option('-c','--sql-create',dest='sqlCreate',default=False,action='store_true',
			  help='Print out an sql create command for the table.')

	addMsgOptions(parser)

	(options,args) = parser.parse_args()
	success=True

	if options.doctest:
		import os; print os.path.basename(sys.argv[0]), 'doctests ...',
		sys.argv= [sys.argv[0]]
		if options.verbose: sys.argv.append('-v')
		import doctest
		numfail,numtests=doctest.testmod()
		if numfail==0: print 'ok'
		else: 
			print 'FAILED'
			success=False

	if not success: sys.exit('Something Failed')
	del success # Hide success from epydoc

	if options.unittest:
		sys.argv = [sys.argv[0]]
		if options.verbose: sys.argv.append('-v')
		unittest.main()

	outfile = sys.stdout
	if None!=options.outputFileName:
		outfile = file(options.outputFileName,'w')


	if options.doEncode:
		# First make sure all non required options are specified
		if None==options.time_monthField: parser.error("missing value for time_monthField")
		if None==options.time_dayField: parser.error("missing value for time_dayField")
		if None==options.time_hourField: parser.error("missing value for time_hourField")
		if None==options.time_minField: parser.error("missing value for time_minField")
		if None==options.stationidField: parser.error("missing value for stationidField")
		if None==options.pos_longitudeField: parser.error("missing value for pos_longitudeField")
		if None==options.pos_latitudeField: parser.error("missing value for pos_latitudeField")
		if None==options.flowField: parser.error("missing value for flowField")
		msgDict={
			'time_month': options.time_monthField,
			'time_day': options.time_dayField,
			'time_hour': options.time_hourField,
			'time_min': options.time_minField,
			'stationid': options.stationidField,
			'pos_longitude': options.pos_longitudeField,
			'pos_latitude': options.pos_latitudeField,
			'flow': options.flowField,
			'reserved': '0',
		}

		bits = encode(msgDict)
		if 'binary'==options.ioType: print str(bits)
		elif 'nmeapayload'==options.ioType:
		        # FIX: figure out if this might be necessary at compile time
			print "bitLen",len(bits)
			bitLen=len(bits)
			if bitLen%6!=0:
			    bits = bits + BitVector(size=(6 - (bitLen%6)))  # Pad out to multiple of 6
			print "result:",binary.bitvectoais6(bits)[0]


		# FIX: Do not emit this option for the binary message payloads.  Does not make sense.
		elif 'nmea'==options.ioType: sys.exit("FIX: need to implement this capability")
		else: sys.exit('ERROR: unknown ioType.  Help!')


	if options.sqlCreate:
		sqlCreateStr(outfile,options.fieldList)

	if options.printCsvfieldList:
		# Make a csv separated list of fields that will be displayed for csv
		if None == options.fieldList: options.fieldList = fieldList
		import StringIO
		buf = StringIO.StringIO()
		for field in options.fieldList:
			buf.write(field+',')
		result = buf.getvalue()
		if result[-1] == ',': print result[:-1]
		else: print result

	if options.doDecode:
		for msg in args:
			bv = None
			if    'binary'    == options.ioType: bv = BitVector(bitstring=msg)
			elif 'nmeapayload'== options.ioType: bv = binary.ais6tobitvec(msg)
			elif 'nmea'       == options.ioType: bv = binary.ais6tobitvec(msg.split(',')[5])
			else: sys.exit('ERROR: unknown ioType.  Help!')

			printFields(decode(bv),out=outfile,format=options.outputType,fieldList=options.fieldList)
